# FILE: link.tcl
#
# An object which implements an event-driven, asynchronous, line-based
# network connection.  It generates an event Readable, when a line
# is available for reading, and queues lines for eventual writing.
# This is the object which actually owns and manages a Tcl socket.
#
# The Tcl event loop must be running for this event-driven object to function.
#

Oc_Class Net_Link {

    # A maximum backlog on the queue, for detecting possible bugs
    common maxQueueLength 100

    # The socket through which we communicate to the host.
    const public variable socket

    # The name of the remote host
    public variable hostname 

    # The remote port
    public variable port 
    
    # Log for logging messages
    public variable log {
        if {[catch {$log Append "Logging messages generated by $this."}]} {
            error "$log is not able to log messages."
        }
        if {[info exists logHandler]} {
            $logHandler Delete
        }
        Oc_EventHandler New logHandler $log Delete [list $this Delete]
        if {$imadelog} {
            $oldlog Delete
        }
        set oldlog $log
    }
    private variable oldlog = Oc_Nop
    private variable logHandler
    private variable imadelog = 0

    # line read from socket
    private variable lineFromSocket

    # queue of messages to write to socket
    private variable putQueue = {}

    # Constructor is a non-blocking operation.  If it doesn't return
    # an error, it will generate a Ready event if and when it has established
    # a socket connection.  Then, it will generate Readable events whenever
    # it has a line to be read (using the Get method).  It will generate a
    # Delete event whenever it is destroyed.
    Constructor { args } {
        eval $this Configure $args
        if {![info exists log]} {
            # Provide a default log
            $this Configure -log [em_log New _ -title "Net_Link $this Log"]
            set imadelog 1
        }
        # Try to open connection to host
        if {![info exists socket]} {
            if {![info exists hostname] || ![info exists port]} {
                error "Missing required options: -hostname and -port"
            }
            #
            # Evaluation of the following socket command blocks while
            # the connection is established (or not).  Once we support
            # communication between different machines, that delay may
            # be intolerable.  We would like to use the -async flag to
            # make the call non-blocking, but a bug in the Tcl library
            # on Windows makes it impossible to detect a failed async
            # connection.  The writable fileevent never fires as the
            # documentation says it should.  There is a patch available
            # for Tcl 8.0.3 which fixes this, but until that makes it
            # into the core, we're stuck.  The patch is at
            # <http://www.scriptics.com/patches/tcl8.0.3/tclUnixChan.c.893.txt>
            #
            if {[catch {socket $hostname $port} msg]} {
                $log Append "Can't connect to $hostname:$port:
	$msg"
                error "Can't connect to $hostname:$port:
	$msg"
            }
            set socket $msg
        }
        after 20 $this SocketOpen
    }

    # handler called when connection to server is established (or not)
    method SocketOpen {} {
        $log Append "$this SocketOpen"
        fileevent $socket writable {}
        # WARNING: Here we're using fconfigure to determine whether or not
        # the socket has established a connection.  However, the ability
        # to determine some value for the -peername option may not
        # correspond to knowing that a connection exists on all platforms.
        if {[catch {fconfigure $socket -peername} peerinfo]} {
            $log Append "No answer on $hostname:$port"
            $this Delete
            return
        }
        set hostname [lindex $peerinfo 1]
        set port [lindex $peerinfo 2]
        fconfigure $socket -blocking 0 -buffering line -translation {auto crlf}
        $log Append "Connected to $hostname:$port"
        $log Append "Peer: $peerinfo"
        # Be sure to flush any queued messages before the application exits
        Oc_EventHandler New _ Oc_Main Exit [list $this Flush] -oneshot 1 \
                -groups $this
        Oc_EventHandler Generate $this Ready
        fileevent $socket readable [list $this Receive]
    }

    # handler called when there is data available from the host
    # It invokes other events describing what it receives.
    method Receive {} {
        if {[catch {gets $socket line} readCount]} {
            $this SocketError $readCount
            return
        }
        if {$readCount < 0} {
            if {[fblocked $socket]} {
                # 'gets' call is blocked until a whole line is available
                return
            }
            if {[eof $socket]} {
                $this SocketEOF
                return
            }
        }
        set lineFromSocket [Oc_Url Unescape $line]
        Oc_EventHandler Generate $this Readable
    }

    private method SocketError { msg } {
        $log Append "socket error in connection $this to $hostname:$port:
	$msg"
        $this Delete
        return
    }

    private method SocketEOF {} {
        fileevent $socket readable {}
	catch { puts $socket "" } ;# For some reason, this seems to
	## help close the network connection on the other end.
	## -mjd, 98-10-09
        $log Append "socket closed by $hostname:$port in $this"
        $this Delete
        return
    }

    method Get {} {
        return $lineFromSocket
    }

    method Put { msg } {
        # Escape the newlines in the string -- a mini version of
        # 'Oc_Url Escape'
        regsub -all % $msg %25 msg
        regsub -all "\n" $msg %0a msg
        lappend putQueue $msg

        # Use event-driven writing of message to socket, if possible.
        # Unfortunately the [fileevent ? writable] command is broken
        # on Windows platforms in pre-8.0 versions of Tcl.  See the
        # file ./fewtest.tcl for more information.
        #
        global tcl_platform
        if {[string match windows $tcl_platform(platform)] \
                && [info tclversion] < 8} {
            $this Send
        } else {
            fileevent $socket writable [list $this Send]
            if {[llength $putQueue] > $maxQueueLength} {
                # will this message ever be seen?
                $log Append "Maximum queue length exceeded in link $this:
	Possibly a socket error?"
                $this Delete
                return
            }
        }
    }

    method Send {} {
        # This won't block, but could leave data backed up in an internal
        # Tcl buffer.  That's unlikely since we're waiting for a writable
        # socket.
        puts $socket [lindex $putQueue 0]
        set putQueue [lrange $putQueue 1 end]
        if {![llength $putQueue]} {
            fileevent $socket writable {}
        }
    }

    method Flush {} {
        # Flush the queue
        while {[llength $putQueue]} {
            $this Send
        }
    }

    Destructor {
        Oc_EventHandler Generate $this Delete
        Oc_EventHandler DeleteGroup $this
        if {[info exists socket]} {
            fileevent $socket readable {}
            catch { $this Flush }
            $log Append "Closing socket $socket in $this"
            close $socket
        }
        if {[info exists log]} {
            $log Append "Deleting $this ..."
            $this Configure -log Oc_Nop
        }
    }

}
