# FILE: log.tcl
#
#	General logging facility.
#
# Last modified on: $Date: 1999/04/26 21:43:16 $
# Last modified by: $Author: dgp $
#
# This file defines the Tcl command Oc_Log for directing messages from
# multiple sources and in multiple categories to multiple handlers for
# recording or display.
#
# This facility was inspired by a similar facility in Pool, a library
# of Tcl routines by Andreas Kupries <a.kupries@westend.com> which is
# available at <URL:http://www.westend.com/%7Ekupries/doc/pool/index.htm>.
#
# DOME: Consider procs for deleting Types and Sources.

proc Oc_ForceNoTkDefaultMessage {} {
    global tcl_platform
    if {[string match windows $tcl_platform(platform)] \
            && [llength [info commands Oc_WindowsMessageBox]]} {
        proc Oc_DefaultMessage {msg type src} {
            global errorInfo errorCode
            set maxLines 20
            set text "[Oc_Main GetInstanceName] $src $type:\n$msg"
	    if {[string compare $type error] == 0} {
                append text "\n-----------\nStack:\n$errorInfo\n-----------\n"
                append text "Additional info: $errorCode"
	    }
            set newtext [join [lrange [split $text \n] 0 $maxLines] \n]
            if {[string compare $text $newtext] != 0} {
                append newtext "\n(message truncated)"
            }
            Oc_WindowsMessageBox $newtext
        }
    } else {
        proc Oc_DefaultMessage {msg type src} {
            global errorInfo errorCode
            set maxLines 20
            set text "[Oc_Main GetInstanceName] $src $type:\n$msg"
	    if {[string compare $type error] == 0} {
                append text "\n-----------\nStack:\n$errorInfo\n-----------\n"
                append text "Additional info: $errorCode"
	    }
            set newtext [join [lrange [split $text \n] 0 $maxLines] \n]
            if {[string compare $text $newtext] != 0} {
                append newtext "\n(message truncated)"
            }
            if {[llength [info commands tcl_puts]] 
                    && [llength [info commands consoleinterp]]} {
                # Bizarre case where Tk is loaded, console
                # is in use, but the app insists it wants
                # messages sent to the launching shell
                # window by default.
                tcl_puts stderr $newtext
            } else {
		puts stderr $newtext
            }
            flush stderr
        }
    }
}

Oc_Class Oc_Log {
    array common db {, ""}

    proc AddType {type} {db} {
        if {[regexp , $type]} {
            return -code error \
                    "The character ',' is not allowed in a message type."
        }
        if {![info exists db(,$type)]} {
            foreach _ [array names db *,] {
                regexp ^(.*),$ $_ _ src
                array set db [list $src,$type $db($src,)]
            }
        }
    }

    proc AddSource {src} {db} {
        if {[regexp , $src]} {
            return -code error \
                    "The character ',' is not allowed in a message source."
        }
        if {![info exists db($src,)]} {
            foreach _ [array names db ,*] {
                regexp ^,(.*)$ $_ _ type 
                array set db [list $src,$type $db(,$type)]
            }
        }
    }

    proc SetLogHandler {hdlr type {src ""}} {db} {
        if {![info exists db($src,$type)]} {
            return -code error "Unknown (source,type) pair: ($src,$type)"
        }
        set db($src,$type) $hdlr
    }

    proc GetLogHandler {type {src ""}} {db} {
        if {![info exists db($src,$type)]} {
            return -code error "Unknown (source,type) pair: ($src,$type)"
        }
        return $db($src,$type)
    }

    proc Log {msg {type ""} {src ""}} {db} {
        if {![info exists db($src,$type)]} {
            return -code error "Unknown (source,type) pair: ($src,$type)"
        }
        if {[string match {} $db($src,$type)]} {
            # No handler, do nothing
            return
        }

        global errorInfo errorCode
	set stack "$errorInfo\n($errorCode)"
        # Arguments quoted by [list] so that uplevel doesn't break them apart.
        if {[catch {
                uplevel #0 $db($src,$type) [list $msg $type $src]
                } ret]} {
            if {[string match Oc_Log $src] && [string match error $type]} {
                # Prevent infinite loops
                # This leaves message unreported.  Alternatives?
                return
            }
	    set msg [join [split $msg \n] \n\t]
	    set ret [join [split $ret \n] \n\t]
	    set stack [join [split $stack \n] \n\t]
	    set lstack [join [split "$errorInfo\n($errorCode)" \n] \n\t]
	    set fmsg "LogHandler '$db($src,$type)' failed\n\treporting message\
                    of type '$type'\n\tfrom source '$src'.\n    Original\
                    message:\n\t$msg\n    LogHandler '$db($src,$type)'\
		    error:\n\t$ret\n    Original stack:\n\t$stack\n   \
		    LogHandler stack:\n\t$lstack"
	    error $fmsg $fmsg [list OC Oc_Log]
        }
    }

    ClassConstructor {
        Oc_Log AddSource Oc_Class
	Oc_Log AddSource Oc_Log

        Oc_Log AddType panic	;# Fatal errors
        Oc_Log AddType error	;# Non-fatal errors reported via Tcl's error
				;# unwinding
        Oc_Log AddType warning	;# Non-fatal errors reported some other way
				;# (like from deep within the C++ code)
        Oc_Log AddType info 	;# A noteworthy, unusual event occurred, 
				;# but not an error.
        Oc_Log AddType status	;# The passing of a routine event
				;# (like an iteration count)
        Oc_Log AddType debug	;# Detailed messages useful only for
				;# development.

        Oc_ForceNoTkDefaultMessage

	foreach s {"" Oc_Class Oc_Log} {
            Oc_Log SetLogHandler Oc_DefaultMessage panic $s
            Oc_Log SetLogHandler Oc_DefaultMessage error $s
            Oc_Log SetLogHandler Oc_DefaultMessage warning $s
            Oc_Log SetLogHandler Oc_DefaultMessage info $s
	}
    }

}

# Redefine bgerror to use Oc_log
proc bgerror {msg} {
    global errorCode errorInfo

    if {[string match OC [lindex $errorCode 0]]} {
        Oc_Log Log $msg error [lindex $errorCode 1]
    } else {
        Oc_Log Log $msg error
    }
}

# Redefine tclLog (Tcl 8.0+) to use Oc_log
proc tclLog {string} {
    Oc_Log Log $string warning
}

# If Tk is loaded, use a better message default reporting routine
if {[Oc_Main HasTk]} {
    proc Oc_DefaultMessage {msg type src} {
	global errorInfo errorCode
	foreach {ei ec} [list $errorInfo $errorCode] {break}
        if {[catch {winfo exists .} result] || !$result} {
            # Main application window '.' has been destroyed.
            Oc_ForceNoTkDefaultMessage
	    foreach {errorInfo errorcode} [list $ei $ec] {break}
            set code [catch {
                    uplevel [list Oc_DefaultMessage $msg $type $src]
                    } result]
            return -code $code $result
        }
        option add *Dialog.msg.wrapLength 6i startupFile
        switch $type {
            info -
            status -
            debug {
                set default Continue
                set options [list Continue]
                set bitmap info
            }
            warning {
                set default Continue
                set options [list Die Stack Continue]
                set bitmap warning
                set stackoptions [list Die Continue]
                set stackdefault Continue
            }
            panic {
                set default Die
                set options [list Die Stack]
                set bitmap error
                set stackoptions [list Die]
                set stackdefault Die
            }
            error -
            default {
                set default Die
                set options [list Die Stack Continue]
                set bitmap error
                set stackoptions [list Die Continue]
                set stackdefault Die
            }
        }
        set i 0
        set switchbody {}
        foreach opt $options {
            if {[string match $opt $default]} {
                set defnum $i
                lappend switchbody -1 -
            }
            lappend switchbody $i
            switch $opt {
                Continue {
                    lappend switchbody {}
                }
                Die {
                    if {[string match panic $type]} {
                        lappend switchbody {}
                    } else {
                        lappend switchbody {exit 1}
                    }
                }
                Stack {
                    lappend switchbody {
                        set si 0
                        set sswitchbody {}
                        foreach opt $stackoptions {
                            if {[string match $opt $stackdefault]} {
                                set sdefnum $si
                                lappend sswitchbody -1 -
                            }
                            lappend sswitchbody $si
                            switch $opt {
                                Continue {
                                    lappend sswitchbody {}
                                }
                                Die {
                                    if {[string match panic $type]} {
                                        lappend sswitchbody {}
                                    } else {
                                        lappend sswitchbody {exit 1}
                                    }
                                }
                            }
                            incr si
                        }
                        switch [eval [list tk_dialog .ocdefaultmessage \
                                "[Oc_Main GetInstanceName] Stack Trace" \
                                "[Oc_Main GetInstanceName] Stack:\
                                 \n$ei\n----------\nAdditional\
                                 info: $ec" info $sdefnum] \
                                 $stackoptions] $sswitchbody
                    }
                }
            }
            incr i
        }
        switch [eval [list tk_dialog .ocdefaultmessage \
                "[Oc_Main GetInstanceName] $src $type:" \
                "[Oc_Main GetInstanceName] $src $type:\n$msg" \
                $bitmap $defnum] $options] $switchbody
    }
}

