/* FILE: mesh.cc                -*-Mode: c++-*-
 *
 * Abstract mesh class and children
 *
 */

#include <math.h>
#include <stdio.h>
#include <string.h>

#include "nb.h"

#include "fileio.h"
#include "mesh.h"

/* End includes */     // This is an optional directive to build.tcl,
                       // that excludes the remainder of the file from
                       // '#include ".*"' dependency building

//////////////////////////////////////////////////////////////////////////
// Abstract Mesh class
void Vf_Mesh::SetFilename(const char* _Filename) {
  if(Filename!=NULL) delete[] Filename;
  if(_Filename==NULL) Filename=NULL;
  else {
    Filename=new char[strlen(_Filename)+1];
    strcpy(Filename,_Filename);
  }
}
void Vf_Mesh::SetTitle(const char* _Title) {
  if(Title!=NULL) delete[] Title;
  if(_Title==NULL) Title=NULL;
  else {
    Title=new char[strlen(_Title)+1];
    strcpy(Title,_Title);
  }
}
void Vf_Mesh::SetDescription(const char* _Description) {
  if(Description!=NULL) delete[] Description;
  if(_Description==NULL) Description=NULL;
  else {
    Description=new char[strlen(_Description)+1];
    strcpy(Description,_Description);
  }
}
void Vf_Mesh::SetValueUnit(const char* _ValueUnit) {
  if(ValueUnit!=NULL) delete[] ValueUnit;
  if(_ValueUnit==NULL) ValueUnit=NULL;
  else {
    ValueUnit=new char[strlen(_ValueUnit)+1];
    strcpy(ValueUnit,_ValueUnit);
  }
}
const char* Vf_Mesh::GetName() const { 
  if(Filename==NULL) return "";
  return Filename; 
}
const char* Vf_Mesh::GetTitle() const {
  if(Title==NULL) return "";
  return Title; 
}
const char* Vf_Mesh::GetDescription() const {
  if(Description==NULL) return "";
  return Description; 
}
const char* Vf_Mesh::GetValueUnit() const {
  if(ValueUnit==NULL) return "";
  return ValueUnit; 
}


//////////////////////////////////////////////////////////////////////////
// Regular 3D grid of Nb_Vec3<REAL4>'s
const ClassDoc Vf_GridVec3f::class_doc("Vf_GridVec3f",
		    "Michael J. Donahue (michael.donahue@nist.gov)",
		    "1.0.0","11-Dec-1996");

// The following constructor should be called from file input functions.
Vf_GridVec3f::Vf_GridVec3f(const char* filename,
                           const char* title,const char* desc,
			   const char* valueunit,REAL8 valuemultiplier,
                           INT4m newi,INT4m newj,INT4m newk,
                           const Nb_Vec3<REAL4> &basept,
                           const Nb_Vec3<REAL4> &gridstep,
                           const Nb_List< Nb_Vec3<REAL4> > &newboundary,
                           const Nb_BoundingBox<REAL4> &newdatarange,
                           const Nb_BoundingBox<REAL4> &newrange)
  :Vf_Mesh("Vf_GridVec3f",filename,title,desc,valueunit,valuemultiplier)
{
  SetSize(newi,newj,newk);
  coords_base=basept;
  coords_step=gridstep;
  if(coords_step.x==0.) coords_step.x=1.;
  if(coords_step.y==0.) coords_step.y=1.;
  if(coords_step.z==0.) coords_step.z=1.;
  boundary=newboundary;
  data_range=newdatarange;
  range=newrange;
}

// Problem <-> Nodal Coordinates Conversion Routines
// The conversion from nodes to problem coordinates is
//    (i,j,k) -> base+(i*step.x,j*step.y,k*step.z).
// NOTE 1: The grid vector access functions, e.g., operator(), access
//   via the integral "nodal" coordinates, *not* in terms of the
//   (real valued) problem coordinates.
// NOTE 2: The values in coords_step are guaranteed to be !=0,
//   but may be negative!  However, GetStdStep() (below), returns
//   a value >0.
// NOTE 3: The coords conversion routines don't preserve order
//   if step.? is negative, e.g., say (min.x,min.y) < (max.x,max.y)
//   going in, but step.x is negative.  Then after conversion,
//   min.y~ will still be < max.y~, but now min.x~>max.x~.

Nb_Vec3<REAL4>
Vf_GridVec3f::NodalToProblemCoords(Nb_Vec3<REAL4> const &w_nodal)
  const
{
  Nb_Vec3<REAL4> v(w_nodal);
  v.x*=coords_step.x;
  v.y*=coords_step.y;
  v.z*=coords_step.z;
  v+=coords_base;
  return v;
}

Nb_Vec3<REAL4>
Vf_GridVec3f::ProblemToNodalCoords(Nb_Vec3<REAL4> const &v_problem)
  const
{
  Nb_Vec3<REAL4> w(v_problem);
  w-=coords_base;
  w.x/=coords_step.x;
  w.y/=coords_step.y;
  w.z/=coords_step.z;
  return w;
}

INT4m Vf_GridVec3f::ColorQuantityTypes(Nb_List<Nb_DString> &types) const
{
  types.Clear();
  types.Append(Nb_DString("x"));
  types.Append(Nb_DString("y"));
  types.Append(Nb_DString("z"));
  types.Append(Nb_DString("xy-angle"));
  types.Append(Nb_DString("div"));
  types.Append(Nb_DString("none"));
  return types.GetSize();
}

REAL4m Vf_GridVec3f::GetVecShade(const char* colorquantity,
                                 const Nb_Vec3<REAL4>& v) const
{
  REAL4m shade = 0.5;
  if(strcmp("xy-angle",colorquantity)==0) {
    if(fabs(v.x)>REAL4_EPSILON || fabs(v.y)>REAL4_EPSILON) {
      shade = atan2(v.y,v.x)*(1./(2*PI));
      if(shade<0) shade+=1.0;
    }
  } else {
    REAL4m mag=v.Mag();   if(mag<REAL4_EPSILON) mag=1.0;
    if(strcmp("x",colorquantity)==0)      shade = (1.+v.x/mag)/2.;
    else if(strcmp("y",colorquantity)==0) shade = (1.+v.y/mag)/2.;
    else if(strcmp("z",colorquantity)==0) shade = (1.+v.z/mag)/2.;
    else                                  shade = -1.;
  }
  return shade;
}

INT4m Vf_GridVec3f::FindClosest(const Nb_Vec3<REAL4> &pos,
				Nb_LocatedVector<REAL4> &lv)
{
  Nb_Vec3<REAL4> npos=ProblemToNodalCoords(pos);
  INT4m i,isize,j,jsize,k,ksize;
  isize=jsize=ksize=1;  // Just to shut up compiler.
  GetDimens(isize,jsize,ksize);
  i=INT4m(OC_ROUND(npos.x));
  if(i<0) i=0;  else if(i>=isize) i=isize-1;
  j=INT4m(OC_ROUND(npos.y));
  if(j<0) j=0;  else if(j>=jsize) j=jsize-1;
  k=INT4m(OC_ROUND(npos.z));
  if(k<0) k=0;  else if(k>=ksize) k=ksize-1;
  lv.value=GridVec(i,j,k);
  lv.location=
    NodalToProblemCoords(Nb_Vec3<REAL4>(REAL4(i),REAL4(j),REAL4(k)));
  return 0;
}

INT4m Vf_GridVec3f::GetDisplayList(REAL4m &xstep_request,
                                   REAL4m &ystep_request,
                                   REAL4m &zstep_request,
                                   const Nb_BoundingBox<REAL4> &dsprange,
                                   const char *colorquantity,
                                   Nb_List<Vf_DisplayVector> &display_list)
{ // Fills display_list with points from mesh inside "dsprange", with
  // step size close to ?step_request, which is reset to value actually
  // used.
  //   Use ?step_request=0 to remove constraint on that axis, e.g.,
  // if xstep_request=ystep_request=zstep_request=0, then this
  // routine places all points inside "dsprange" into display_list.
  //   NOTE: Non-integral values for step_request lead to unattractive
  // banding in the output display list, and so are currently
  // overridden.
#define MEMBERNAME "GetDisplayList"
  if(xstep_request<0 || ystep_request<0 || zstep_request<0)
    FatalError(-1,STDDOC,"%s: xstep=%f, ystep=%f, zstep=%d",
	       ErrBadParam,xstep_request,ystep_request,zstep_request);

  enum Gv3Color { GV3_NONE, GV3_X, GV3_Y, GV3_Z, GV3_XYANGLE, GV3_DIV };
  Gv3Color cq=GV3_NONE;
  if(colorquantity!=NULL) {
    if(strcmp("x",colorquantity)==0)             cq=GV3_X;
    else if(strcmp("y",colorquantity)==0)        cq=GV3_Y;
    else if(strcmp("z",colorquantity)==0)        cq=GV3_Z;
    else if(strcmp("xy-angle",colorquantity)==0) cq=GV3_XYANGLE;
    else if(strcmp("div",colorquantity)==0)      cq=GV3_DIV;
  }

  display_list.Clear();
  INT4m isize,jsize,ksize;
  GetDimens(isize,jsize,ksize);
  if(isize<1 || jsize<1 || ksize<1)
    return display_list.GetSize(); // Empty grid
  Nb_Vec3<REAL4> rangemin,rangemax;
  dsprange.GetExtremes(rangemin,rangemax);

  // Convert dsprange & step requests from problem coords to nodal coords
  rangemin=ProblemToNodalCoords(rangemin);
  rangemax=ProblemToNodalCoords(rangemax);
  if(coords_step.x<0) swap(rangemin.x,rangemax.x); // Adjust for ordering
  if(coords_step.y<0) swap(rangemin.y,rangemax.y); // across coord transform.
  if(coords_step.z<0) swap(rangemin.z,rangemax.z);

  REAL4m nxstep=xstep_request/fabs(coords_step.x);
  REAL4m nystep=ystep_request/fabs(coords_step.y);
  REAL4m nzstep=zstep_request/fabs(coords_step.z);

  INT4m ifirst,ilast,jfirst,jlast,kfirst,klast;
  ifirst=INT4m(ceil(rangemin.x)); ilast=INT4m(floor(rangemax.x))+1;
  jfirst=INT4m(ceil(rangemin.y)); jlast=INT4m(floor(rangemax.y))+1;
  kfirst=INT4m(ceil(rangemin.z)); klast=INT4m(floor(rangemax.z))+1;

  if(ifirst<0 || ilast>isize || jfirst<0 || jlast>jsize ||
     kfirst<0 || klast>ksize) {
#if DEBUGLVL > 999
    NonFatalError(STDDOC,"Range box out-of-bounds: (%d,%d,%d) "
		  "to (%d,%d,%d), not in (0,0,0) to (%d,%d,%d)",
		  ifirst,jfirst,kfirst,ilast,jlast,klast,
		  isize,jsize,ksize);
#endif
    if(ifirst<0) ifirst=0; if(ilast>isize) ilast=isize;
    if(jfirst<0) jfirst=0; if(jlast>jsize) jlast=jsize;
    if(kfirst<0) kfirst=0; if(klast>ksize) klast=ksize;
  }

  // Calculate step sizes
  // Note: Non-integer step size produce ugly banding in output sampling
  nxstep=OC_ROUND(nxstep);
  nystep=OC_ROUND(nystep);
  nzstep=OC_ROUND(nzstep);

  REAL4m max_step=REAL4m(OC_MAX(OC_MAX(isize,jsize),ksize));
  if(nxstep>max_step) nxstep=max_step;
  if(nystep>max_step) nystep=max_step;
  if(nzstep>max_step) nzstep=max_step;

  INT4m istep=INT4m(nxstep);
  INT4m jstep=INT4m(nystep);
  INT4m kstep=INT4m(nzstep);

  if(istep<1) istep=1;  if(jstep<1) jstep=1;  if(kstep<1) kstep=1;

  // Determine offsets
  INT4m ioff=((ilast-ifirst-1)%istep)/2;
  INT4m joff=((jlast-jfirst-1)%jstep)/2;
  INT4m koff=((klast-kfirst-1)%kstep)/2;

  INT4m i,j,k;
  INT4m istart,jstart,kstart;
  Nb_Vec3<REAL4> position;  // Position, in "Problem" coords
  REAL4m xstart,ystart,zstart,xstep,ystep,zstep;

  istart=ifirst+ioff;
  xstart=coords_base.x+coords_step.x*REAL4m(istart);
  xstep=coords_step.x*REAL4m(istep);

  jstart=jfirst+joff;
  ystart=coords_base.y+coords_step.y*REAL4m(jstart);
  ystep=coords_step.y*REAL4m(jstep);

  kstart=kfirst+koff;
  zstart=coords_base.z+coords_step.z*REAL4m(kstart);
  zstep=coords_step.z*REAL4m(kstep);

  REAL4m shade,max_shade=-1.,mag;
  Nb_Vec3<REAL4> v;

  REAL4m xdivscale=1.;  // Scaling for divergence calculation
  if(xstep!=0.) { xdivscale=ystep/xstep; } // Note that divergence
  // is scaled and translated in the end so that 0 div => 0.5 shade,
  // and max(|div|) is either 0.0 or 1.0 shade.

  for(k=kstart,position.z=zstart;k<klast;k+=kstep,position.z+=zstep) {
    for(j=jstart,position.y=ystart;j<jlast;j+=jstep,position.y+=ystep) {
      for(i=istart,position.x=xstart;i<ilast;i+=istep,position.x+=xstep) {
	v=GridVec(i,j,k);
	shade=0.5;
	switch(cq) {
	case GV3_X:
	  if((mag=v.Mag())>REAL4_EPSILON) shade=(1.+v.x/mag)/2.; break;
	case GV3_Y:
	  if((mag=v.Mag())>REAL4_EPSILON) shade=(1.+v.y/mag)/2.; break;
	case GV3_Z:
	  if((mag=v.Mag())>REAL4_EPSILON) shade=(1.+v.z/mag)/2.; break;
	case GV3_XYANGLE:
	  if(fabs(v.x)>REAL4_EPSILON || fabs(v.y)>REAL4_EPSILON) {
	    shade = atan2(v.y,v.x)*(1./(2*PI));
	    if(shade<0) shade+=1.0;
	  }
	  break;
        case GV3_DIV:
	  shade=0.;
	  if(i+istep<isize) shade+=GridVec(i+istep,j,k).x;
	  if(i-istep>=0)    shade-=GridVec(i-istep,j,k).x;
	  shade*=xdivscale;
	  if(j+jstep<jsize) shade+=GridVec(i,j+jstep,k).y;
	  if(j-jstep>=0)    shade-=GridVec(i,j-jstep,k).y;
          if(fabs(shade)>max_shade) max_shade=fabs(shade);
	default:
	  break;
	}
	display_list.Append(Vf_DisplayVector(position,v,shade));
      }
    }
  }

  if(cq==GV3_DIV) {
    // Normalize all shading
    REAL4m mult=0.;
    if(max_shade>0.) mult=1./(2.*max_shade);
    for(Vf_DisplayVector* dvp=display_list.GetFirst(); dvp!=NULL;
        dvp=display_list.GetNext()) {
      dvp->shade=0.5+dvp->shade*mult;
    }
  }

  // Export actual step sizes, in problem units
  xstep_request=nxstep*fabs(coords_step.x);
  ystep_request=nystep*fabs(coords_step.y);
  zstep_request=nzstep*fabs(coords_step.z);

  return display_list.GetSize();
#undef MEMBERNAME
}

//////////////////////////////////////////////////////////////////////////
// General (non-regular) 3D mesh of Nb_Vec3<REAL4>'s
const ClassDoc Vf_GeneralMesh3f::class_doc("Vf_GeneralMesh3f",
		    "Michael J. Donahue (michael.donahue@nist.gov)",
		    "1.0.0","1-Sep-1997");

// Generic Mesh functions /////////////////////

INT4m Vf_GeneralMesh3f::ColorQuantityTypes(Nb_List<Nb_DString> &types) const
{
  types.Clear();
  types.Append(Nb_DString("x"));
  types.Append(Nb_DString("y"));
  types.Append(Nb_DString("z"));
  types.Append(Nb_DString("xy-angle"));
  types.Append(Nb_DString("none"));
  return types.GetSize();
}

REAL4m Vf_GeneralMesh3f::GetVecShade(const char* colorquantity,
                                     const Nb_Vec3<REAL4>& v) const
{
  REAL4m shade = 0.5;
  if(strcmp("xy-angle",colorquantity)==0) {
    if(fabs(v.x)>REAL4_EPSILON || fabs(v.y)>REAL4_EPSILON) {
      shade = atan2(v.y,v.x)*(1./(2*PI));
      if(shade<0) shade+=1.0;
    }
  } else {
    REAL4m mag=v.Mag();   if(mag<REAL4_EPSILON) mag=1.0;
    if(strcmp("x",colorquantity)==0)      shade = (1.+v.x/mag)/2.;
    else if(strcmp("y",colorquantity)==0) shade = (1.+v.y/mag)/2.;
    else if(strcmp("z",colorquantity)==0) shade = (1.+v.z/mag)/2.;
    else                                  shade = -1.;
  }
  return shade;
}

INT4m Vf_GeneralMesh3f::FindClosest(const Nb_Vec3<REAL4> &pos,
				    Nb_LocatedVector<REAL4> &lv)
{
  INT4m dummy;
  Nb_Vec3<REAL4> work_pos=pos;
  Nb_LocatedVector<REAL4> *lvp;
  data_range.MoveInto(work_pos);
  INT4m errorcode=mesh.GetClosest2D(work_pos,lvp,dummy);
  if(errorcode==0) lv=*lvp;
  return errorcode;
}

INT4m
Vf_GeneralMesh3f::GetDisplayList(REAL4m &xstep_request,
                                 REAL4m &ystep_request,
                                 REAL4m & /* zstep_request */,
                                 const Nb_BoundingBox<REAL4> & range_request,
                                 const char *colorquantity,
                                 Nb_List<Vf_DisplayVector> &display_list)
{
#define MEMBERNAME "GetDisplayList"
  enum Gm3Color { GM3_NONE, GM3_X, GM3_Y, GM3_Z, GM3_XYANGLE };
  Gm3Color cq=GM3_NONE;
  if(colorquantity!=NULL) {
    if(strcmp("x",colorquantity)==0)      cq=GM3_X;
    else if(strcmp("y",colorquantity)==0) cq=GM3_Y;
    else if(strcmp("z",colorquantity)==0) cq=GM3_Z;
    else if(strcmp("xy-angle",colorquantity)==0) cq=GM3_XYANGLE;
  }
  REAL4m shade,mag;

  display_list.Clear();
  mesh.ClearSelectCounts();
  if(xstep_request<SQRT_REAL4_EPSILON || 
     ystep_request<SQRT_REAL4_EPSILON) {
    // Return whole list
    Nb_LocatedVector<REAL4> lv;
    mesh.ResetWholeAccess();
    while(mesh.GetWholeNext(lv)==0) {
	shade=0.5;
	switch(cq) {
	case GM3_X:
	  if((mag=lv.value.Mag())>REAL4_EPSILON)
	    shade=(1.+lv.value.x/mag)/2.;
	  break;
	case GM3_Y:
	  if((mag=lv.value.Mag())>REAL4_EPSILON)
	    shade=(1.+lv.value.y/mag)/2.;
	  break;
	case GM3_Z:
	  if((mag=lv.value.Mag())>REAL4_EPSILON)
	    shade=(1.+lv.value.z/mag)/2.;
	  break;
	case GM3_XYANGLE:
	  if(fabs(lv.value.x)>REAL4_EPSILON ||
	     fabs(lv.value.y)>REAL4_EPSILON) {
	    shade = atan2(lv.value.y,lv.value.x)*(1./(2*PI));
	    if(shade<0) shade+=1.0;
	  }
	  break;
	default:
	  break;
	}
	display_list.Append(Vf_DisplayVector(lv.location,lv.value,shade));
    }
  }
  else {
    // Otherwise, center a maximal regular grid, with lattice
    // steps xstep_request, ystep_request, inside working mesh box
    // region "work_range", and then use mesh.GetClosest2D to collect
    // samples from the mesh closest to the overlaid regular grid.
    Nb_BoundingBox<REAL4> work_range(data_range);
    work_range.IntersectWith(range_request);
    Nb_Vec3<REAL4> minpt,maxpt;
    work_range.GetExtremes(minpt,maxpt);
    const REAL4m fudge=1.001;  // Fudge factor, used to help
    /// against display "banding" caused by rounding.

    REAL4m xextent=maxpt.x-minpt.x;
    REAL4m xstep=xstep_request;
    INT4m  ixcount=int(floor((xextent*fudge)/xstep))+1;
    REAL4m xmargin=(xextent-xstep*(ixcount-1))/(2.*fudge);

    REAL4m yextent=maxpt.y-minpt.y;
    REAL4m ystep=ystep_request;
    INT4m  iycount=int(floor((yextent*fudge)/ystep))+1;
    REAL4m ymargin=(yextent-ystep*(iycount-1))/(2.*fudge);

    // If margin is close to an integral multiple of step_hints,
    // then round it to the multiple.  This should help remove
    // some display "banding" when sampling from grids that are
    // in actual fact regular.
    if(step_hints.x>REAL4_EPSILON && xmargin>REAL4_EPSILON) {
      REAL4m mult=xmargin/step_hints.x;
      REAL4m whole=OC_ROUND(mult);
      REAL4m frac=mult-whole;
      if(whole>REAL4_EPSILON && fabs(frac)<0.1)
	xmargin=step_hints.x*whole;
    }
    if(step_hints.y>REAL4_EPSILON && ymargin>REAL4_EPSILON) {
      REAL4m mult=ymargin/step_hints.y;
      REAL4m whole=OC_ROUND(mult);
      REAL4m frac=mult-whole;
      if(whole>REAL4_EPSILON && fabs(frac)<0.1)
	ymargin=step_hints.y*whole;
    }

    int ix,iy; REAL4m x,y;
    INT4m select_count;
    Nb_LocatedVector<REAL4> *lvp;
    Nb_Vec3<REAL4> pos(0.,0.,0.);
    for(ix=0, x=minpt.x+xmargin; ix<ixcount; ix++, x+=xstep) {
      pos.x=x;
      for(iy=0, y=minpt.y+ymargin; iy<iycount; iy++, y+=ystep) {
	pos.y=y;
	mesh.GetClosest2D(pos,lvp,select_count);
	if(select_count==1) {
	  shade=0.5;
	  switch(cq) {
	  case GM3_X:
	    if((mag=lvp->value.Mag())>REAL4_EPSILON)
	      shade=(1.+lvp->value.x/mag)/2.;
	    break;
	  case GM3_Y:
	    if((mag=lvp->value.Mag())>REAL4_EPSILON)
	      shade=(1.+lvp->value.y/mag)/2.;
	    break;
	  case GM3_Z:
	    if((mag=lvp->value.Mag())>REAL4_EPSILON)
	      shade=(1.+lvp->value.z/mag)/2.;
	    break;
	  default:
	    break;
	  }
	  display_list.Append(Vf_DisplayVector(lvp->location,
                                               lvp->value,shade));
	}
      }
    }
  }
  return display_list.GetSize();
#undef MEMBERNAME
}

void Vf_GeneralMesh3f::SetApproximateStepHints()
{ // We should have a sophisticated algorithm here that computes
  // and returns the average nearest neighbor distance, but for
  // now we'll just calculate some "average" cell dimensions.
  // NOTE: This functions overwrites any pre-existing value in
  //       member variable step_hints, which may override a file
  //       specified value.  A client may want to check to see
  //       if step_hints.x==0 before calling this routine.
#define MEMBERNAME "SetApproximateStepHints"
  step_hints.Set(0.,0.,0.);
  INT4m point_count=GetSize();
  if(point_count<1) return;  // Meaningless case
  REAL4m total_area=fabs(data_range.GetWidth()*data_range.GetHeight());
  /// fabs() for safety.

  REAL4m avecell_area=total_area/point_count;  // "Cells" here
  /// really mean "cells" for display purposes, which are centered
  /// about the sample nodes.

  step_hints.x=sqrt(avecell_area);
  step_hints.y=step_hints.x;
  step_hints.z=0.;
  return;
#undef MEMBERNAME
}

void Vf_GeneralMesh3f::SetBoundaryList(const Nb_List< Nb_Vec3<REAL4> >
                                       &boundary_list)
{
#define MEMBERNAME "SetBoundaryList"
  boundary=boundary_list;
  for(Nb_Vec3<REAL4> *vp=boundary.GetFirst();vp!=NULL;
      vp=boundary.GetNext()) {
    range.ExpandWith(*vp);
  }
#undef MEMBERNAME
}

REAL4m Vf_GeneralMesh3f::GetApproximateCellWidth() const
{ // Uses step_hints.  If step_hints.x==0, then first
  // calls SetApproximateStepHints.
  // NOTE: This function is "conceptually const" as opposed to
  //  "bitwise const"; cf. p76-78 of Meyer's book, "Effective C++."
  //  As a result, we need to do some ugly un-const casting.
  //  This is dangerous, since it looses compiler checks against
  //  changes to ::SetApproximateStepHints that may destroy
  //  conceptual const-ness of this function.
#define MEMBERNAME "GetApproximateCellWidth"
  Vf_GeneralMesh3f * const localThis = (Vf_GeneralMesh3f *)this;
  if(step_hints.x==0) localThis->SetApproximateStepHints();
  if(step_hints.x==0) return 1.0;  // Nonsense case
  return sqrt((OC_SQ(step_hints.x)+OC_SQ(step_hints.y))/2.);
  /// Using the diagonal instead of a simple average of step_hints.x
  /// and step_hints.y is more robust against non-square elements.
#undef MEMBERNAME
}

Nb_Vec3<REAL4> Vf_GeneralMesh3f::GetApproximateCellDimensions() const
{
  // See note in :::GetApproximateCellWidth() about
  // "conceptually const".
#define MEMBERNAME "GetApproximateCellDimensions"
  Vf_GeneralMesh3f * const localThis = (Vf_GeneralMesh3f *)this;
  if(step_hints.x==0) localThis->SetApproximateStepHints();
  if(step_hints.x==0) return Nb_Vec3<REAL4>(1.,1.,1.);  // Nonsense case
  return step_hints;
#undef MEMBERNAME
}

// Vf_GeneralMesh3f specific functions
void Vf_GeneralMesh3f::ResetMesh()
{
#define MEMBERNAME "ResetMesh"
  mesh.ClearList();
  data_range.Reset();
  boundary.Clear();
  range.Reset();
  step_hints.Set(0.,0.,0.);
#undef MEMBERNAME
}

void Vf_GeneralMesh3f::ResetMesh(const char* newFilename)
{
#define MEMBERNAME "ResetMesh(const char*)"
  SetFilename(newFilename);
  ResetMesh();
#undef MEMBERNAME
}

void Vf_GeneralMesh3f::ResetMesh(const char* newFilename,
                                 const char* newTitle,
                                 const char* newDescription)
{
#define MEMBERNAME "ResetMesh(const char*,const char*)"
  SetTitle(newTitle);
  SetDescription(newDescription);
  ResetMesh(newFilename);
#undef MEMBERNAME
}

void Vf_GeneralMesh3f::AddPoint(const Nb_LocatedVector<REAL4>& lv)
{
#define MEMBERNAME "AddPoint"
  mesh.AddPoint(lv);
  data_range.ExpandWith(lv.location);
  range.ExpandWith(lv.location);
#undef MEMBERNAME
}

void Vf_GeneralMesh3f::SortPoints()
{
#define MEMBERNAME "SortPoints"
  // Guess at cell dimensions, if necessary
  if(step_hints.x==0) SetApproximateStepHints();

  // Make BoxList mesh bounding box big enough so that
  // refinement boxes will include all sampling points
  // we may ask of it.
  if(!range.IsEmpty()) {
    Nb_Vec3<REAL4> min,max;
    range.GetExtremes(min,max);
    mesh.ExpandBoxRegion(min.x,min.y,max.x,max.y);
  }
  mesh.InflateBoxRegion(1.2,1.2);   // Add a 10% margin on all sides

  // OK, now do refinement
  // mesh.Refine(6);
  mesh.Refine(1024,2.,15.);

  ClassDebugMessage(STDDOC,"Mesh refinement stats:\n"
	  "  Number of subboxes: %d\n"
	  "  Average number of points per box: %.2f\n"
	  "  Average box list length:  %.2f\n"
	  "  Memory waste: %d\n",
	  mesh.GetBoxCount(),mesh.GetAveInCount(),mesh.GetAveListCount(),
	  mesh.GetSpaceWaste());

#undef MEMBERNAME
}


